# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Implementation of the Keras API, the high-level API of TensorFlow.

Detailed documentation and user guides are available at
[keras.io](https://keras.io).

"""

import sys as _sys

from keras import __version__
from keras.api._v1.keras import __internal__
from keras.api._v1.keras import activations
from keras.api._v1.keras import applications
from keras.api._v1.keras import backend
from keras.api._v1.keras import callbacks
from keras.api._v1.keras import constraints
from keras.api._v1.keras import datasets
from keras.api._v1.keras import estimator
from keras.api._v1.keras import experimental
from keras.api._v1.keras import initializers
from keras.api._v1.keras import layers
from keras.api._v1.keras import losses
from keras.api._v1.keras import metrics
from keras.api._v1.keras import mixed_precision
from keras.api._v1.keras import models
from keras.api._v1.keras import optimizers
from keras.api._v1.keras import preprocessing
from keras.api._v1.keras import regularizers
from keras.api._v1.keras import utils
from keras.api._v1.keras import wrappers
from keras.engine.input_layer import Input
from keras.engine.sequential import Sequential
from keras.engine.training import Model
from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "keras", public_apis=None, deprecation=True,
      has_lite=False)
